# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from dataclasses import InitVar
from typing import Any

from pydantic import AnyUrl, EmailStr, Field
from pydantic.dataclasses import dataclass
from requests import Response

from ikigai.client.api import ComponentAPI
from ikigai.client.session import Session
from ikigai.utils.compatibility import HTTPMethod

logger = logging.getLogger("ikigai.client")


@dataclass
class Client:
    # Init only vars
    user_email: InitVar[EmailStr]
    api_key: InitVar[str]
    base_url: InitVar[AnyUrl]

    __session: Session = Field(init=False)
    __component_api: ComponentAPI = Field(init=False)

    def __post_init__(
        self, user_email: EmailStr, api_key: str, base_url: AnyUrl
    ) -> None:
        self.__session = Session(
            user_email=user_email, api_key=api_key, base_url=base_url
        )
        self.__component_api = ComponentAPI(session=self.__session)

    def get(self, path: str, params: dict[str, Any] | None = None) -> Response:
        return self.__session.request(method=HTTPMethod.GET, path=path, params=params)

    def post(self, path: str, json: dict[Any, Any] | None = None) -> Response:
        return self.__session.request(method=HTTPMethod.POST, path=path, json=json)

    # APIs

    @property
    def component(self) -> ComponentAPI:
        return self.__component_api



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any, TypedDict

from ikigai.typing.protocol.directory import DirectoryDict
from ikigai.utils.compatibility import NotRequired


class FlowDict(TypedDict):
    project_id: str
    pipeline_id: str
    name: str
    directory: DirectoryDict
    definition: FlowDefinitionDict
    trigger_downstream_pipelines: bool
    high_volume_preference: bool
    schedule: dict
    last_run: NotRequired[dict]
    next_run: NotRequired[dict]
    created_at: str
    modified_at: str


class FlowDefinitionDict(TypedDict):
    facets: list[FacetDict]
    arrows: list[ArrowDict]
    arguments: NotRequired[dict]
    variables: NotRequired[dict[str, FlowVariableDict]]
    model_variables: NotRequired[dict[str, FlowModelVariableDict]]


class FacetDict(TypedDict):
    facet_id: str
    facet_uid: str
    name: NotRequired[str]
    arguments: NotRequired[dict]


class ArrowDict(TypedDict):
    source: str
    destination: str
    arguments: NotRequired[dict]


class FlowVariableDict(TypedDict):
    name: str
    value: Any
    facet_name: NotRequired[str]
    type: str
    is_list: bool


class FlowModelVariableDict(TypedDict):
    facet_name: str
    model_name: str
    model_version: NotRequired[str]
    model_argument_type: str
    model_arguments: list[dict]


class FlowStatusReportDict(TypedDict):
    status: str
    progress: NotRequired[int]
    message: str


class FlowLogDict(TypedDict):
    log_id: str
    status: str
    user: str
    erroneous_facet_id: NotRequired[str]
    message: str
    timestamp: str



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from dataclasses import InitVar
from http import HTTPStatus
from typing import Any

import requests
from pydantic import AnyUrl, ConfigDict, EmailStr, Field
from pydantic.dataclasses import dataclass
from requests import Response

from ikigai.utils.compatibility import HTTPMethod

logger = logging.getLogger("ikigai.client")


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class Session:
    # Init only vars
    user_email: InitVar[EmailStr]
    api_key: InitVar[str]

    base_url: AnyUrl
    __session: requests.Session = Field(init=False)

    def __post_init__(self, user_email: EmailStr, api_key: str) -> None:
        self.__session = requests.Session()
        self.__session.headers.update({"user": user_email, "api-key": api_key})

    def request(
        self,
        method: HTTPMethod,
        path: str,
        params: dict[str, str] | None = None,
        json: dict | None = None,
    ) -> Response:
        logger.debug(
            "[%(method)s] %(path)s %(params)s\n" "json: %(json)s",
            {"method": method, "path": path, "params": params, "json": json},
        )
        url = f"{self.base_url}{path}"
        resp = self.__session.request(
            method=method,
            url=url,
            params=params,
            json=json,
        )
        if resp.status_code < HTTPStatus.BAD_REQUEST:
            return resp
        elif resp.status_code < HTTPStatus.INTERNAL_SERVER_ERROR:
            # A 4XX error happened
            logger.error(
                "Request"
                "[%(method)s] %(path)s %(params)s\n"
                "%(request)s\n\n"
                "Response [%(status)s]"
                "headers: %(response_headers)s\n"
                "%(response)s\n\n",
                {
                    "method": method,
                    "path": path,
                    "params": params,
                    "request": resp.request.body,
                    "status": resp.status_code,
                    "response_headers": resp.headers,
                    "response": resp.text,
                },
            )
            todo = "TODO: Add error reporting"
            raise NotImplementedError(todo)
        else:
            # A 5XX error happened
            logger.error(
                "Request"
                "[%(method)s] %(path)s %(params)s\n"
                "%(request)s\n\n"
                "Response [%(status)s]"
                "headers: %(response_headers)s\n"
                "%(response)s\n\n",
                {
                    "method": method,
                    "path": path,
                    "params": params,
                    "request": resp.request.body,
                    "status": resp.status_code,
                    "response_headers": resp.headers,
                    "response": resp.text,
                },
            )
            todo = "TODO: Add error reporting"
            raise NotImplementedError(todo)
        return resp

    def get(self, path: str, params: dict[str, Any] | None = None) -> Response:
        return self.request(method=HTTPMethod.GET, path=path, params=params)

    def post(self, path: str, json: dict[Any, Any] | None = None) -> Response:
        return self.request(method=HTTPMethod.POST, path=path, json=json)

    def __del__(self) -> None:
        self.__session.close()



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TypedDict

from ikigai.typing.protocol.directory import DirectoryDict


class AppDict(TypedDict):
    project_id: str
    name: str
    owner: str
    description: str
    icon: str
    images: list[str]
    directory: DirectoryDict
    created_at: str
    modified_at: str
    last_used_at: str



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Annotated, Any, Optional

from pydantic import BeforeValidator


def __optional_str(value: Any) -> str | None:
    if not value:
        return None
    return str(value)


OptionalStr = Annotated[Optional[str], BeforeValidator(__optional_str)]



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import time
from collections.abc import Mapping
from datetime import datetime
from enum import Enum
from typing import Any

from pydantic import AliasChoices, BaseModel, EmailStr, Field
from tqdm.auto import tqdm

from ikigai.client import Client
from ikigai.typing.protocol import (
    Directory,
    DirectoryType,
    FlowDefinitionDict,
    FlowDict,
    NamedDirectoryDict,
)
from ikigai.utils.compatibility import UTC, Self
from ikigai.utils.custom_validators import OptionalStr
from ikigai.utils.named_mapping import NamedMapping
from ikigai.utils.shim import flow_versioning_shim

logger = logging.getLogger("ikigai.components")


class FlowDefinition(BaseModel):
    _facets: list = []
    _arrows: list = []
    _arguments: dict = {}
    _variables: dict = {}
    _model_variables: dict = {}

    def to_dict(self) -> FlowDefinitionDict:
        # TODO: Update implementation when feature is available
        return {
            "facets": self._facets,
            "arrows": self._arrows,
            "arguments": self._arguments,
            "variables": self._variables,
            "model_variables": self._model_variables,
        }


class FlowBuilder:
    _app_id: str
    _name: str
    _directory: Directory | None
    _flow_definition: FlowDefinitionDict
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._directory = None
        self._flow_definition = FlowDefinition().to_dict()

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def definition(
        self, definition: Flow | FlowDefinition | FlowDefinitionDict
    ) -> Self:
        if isinstance(definition, FlowDefinition):
            self._flow_definition = definition.to_dict()
            return self

        if isinstance(definition, Flow):
            if definition.app_id != self._app_id:
                error_msg = (
                    "Building flow from a diferent app is not supported\n"
                    "source_app != destination_app "
                    f"({definition.app_id} != {self._app_id})"
                )
                raise ValueError(error_msg)
            flow_dict = self.__client.component.get_flow(flow_id=definition.flow_id)
            self._flow_definition = flow_dict["definition"]
            return self

        if isinstance(definition, dict):
            self._flow_definition = definition
            return self

        error_msg = (
            f"Definition was of type {type(definition)} but, "
            "must be a Flow or FlowDefinition or FlowDefinitionDict"
        )
        raise TypeError(error_msg)

    def directory(self, directory: Directory) -> Self:
        self._directory = directory
        return self

    def build(self) -> Flow:
        flow_id = self.__client.component.create_flow(
            app_id=self._app_id,
            name=self._name,
            directory=self._directory,
            flow_definition=self._flow_definition,
        )

        # Populate Flow object
        flow_dict = self.__client.component.get_flow(flow_id=flow_id)
        flow = Flow.from_dict(data=flow_dict, client=self.__client)
        return flow


class FlowStatus(str, Enum):
    SCHEDULED = "SCHEDULED"
    RUNNING = "RUNNING"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"
    FAILED = "FAILED"
    IDLE = "IDLE"
    UNKNOWN = "UNKNOWN"
    SUCCESS = "SUCCESS"  # Not available via /component/is-pipeline-running

    def __repr__(self) -> str:
        return self.value


class FlowStatusReport(BaseModel):
    status: FlowStatus
    progress: int | None = Field(default=None)
    message: str

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Self:
        self = cls.model_validate(data)
        return self


class RunLog(BaseModel):
    log_id: str
    status: FlowStatus
    user: EmailStr
    erroneous_facet_id: OptionalStr
    data: str = Field(validation_alias="message")
    timestamp: datetime

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Self:
        self = cls.model_validate(data)
        return self


class Flow(BaseModel):
    app_id: str = Field(validation_alias=AliasChoices("app_id", "project_id"))
    flow_id: str = Field(validation_alias=AliasChoices("flow_id", "pipeline_id"))
    name: str
    created_at: datetime
    modified_at: datetime
    __client: Client

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> dict:
        return {
            "flow_id": self.flow_id,
            "name": self.name,
            "created_at": self.created_at,
            "modified_at": self.modified_at,
        }

    def delete(self) -> None:
        self.__client.component.delete_flow(app_id=self.app_id, flow_id=self.flow_id)
        return None

    def rename(self, name: str) -> Self:
        self.__client.component.edit_flow(
            app_id=self.app_id, flow_id=self.flow_id, name=name
        )
        # TODO: handle error case, currently it is a raise NotImplemented from Session
        self.name = name
        return self

    def move(self, directory: Directory) -> Self:
        self.__client.component.edit_flow(
            app_id=self.app_id, flow_id=self.flow_id, directory=directory
        )
        return self

    def status(self) -> FlowStatusReport:
        resp = self.__client.component.is_flow_runing(
            app_id=self.app_id, flow_id=self.flow_id
        )
        return FlowStatusReport.from_dict(resp)

    def run_logs(
        self, max_count: int = 1, since: datetime | None = None
    ) -> list[RunLog]:
        log_dicts = self.__client.component.get_flow_log(
            app_id=self.app_id, flow_id=self.flow_id, max_count=max_count
        )

        run_logs = [RunLog.from_dict(data=log) for log in log_dicts]
        if since is not None:
            run_logs = [log for log in run_logs if log.timestamp > since]
        return run_logs

    def run(self) -> RunLog:
        # Start running pipeline
        self.__client.component.run_flow(app_id=self.app_id, flow_id=self.flow_id)

        return self.__await_run()

    def describe(self) -> FlowDict:
        flow = self.__client.component.get_flow(flow_id=self.flow_id)
        # Apply flow_versioning_shim to allow migration of older flows
        # TODO: Remove this shim after "important" flows are migrated
        facet_specs = self.__client.get(
            path="/component/get-facet-specs",
        ).json()
        shimed_flow = flow_versioning_shim(flow=flow, facet_specs=facet_specs)
        return shimed_flow

    def __await_run(self) -> RunLog:
        start_time = datetime.now(UTC)
        # TODO: Switch to using websockets once they are available
        with tqdm(total=100, dynamic_ncols=True) as progress_bar:
            status_report = self.status()
            progress_bar.desc = status_report.status
            progress_bar.update(0)

            # Initially wait while pipeline is scheduled
            while status_report.status == FlowStatus.SCHEDULED:
                time.sleep(5)
                status_report = self.status()

            last_progress = status_report.progress if status_report.progress else 0
            progress_bar.desc = status_report.status
            progress_bar.update(last_progress)

            # Wait while pipeline is running
            running_states = (
                FlowStatus.RUNNING,
                FlowStatus.STOPPING,
                FlowStatus.UNKNOWN,
            )
            while status_report.status in running_states:
                time.sleep(1)
                status_report = self.status()
                progress = status_report.progress if status_report.progress else 100
                progress_bar.desc = status_report.status
                new_progress = last_progress + max(progress - last_progress, 0)
                progress_bar.update(new_progress - last_progress)
                last_progress = new_progress
            # Flow run completed

            # Get status from logs and update progress bar
            run_logs = self.run_logs(max_count=1, since=start_time)
            if not run_logs:
                # TODO: Give a better error message
                error_msg = (
                    "No logs found for"
                    f" <Flow(flow_id={self.flow_id}, name={self.name})>"
                    f" after the flow started running ({start_time=})."
                )
                raise RuntimeError(error_msg)
            run_log = run_logs[0]

            progress = 100
            progress_bar.desc = run_log.status
            progress_bar.update(progress - last_progress)

            return run_log


class FlowDirectoryBuilder:
    _app_id: str
    _name: str
    _parent: Directory | None
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._parent = None

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def parent(self, parent: Directory) -> Self:
        self._parent = parent
        return self

    def build(self) -> FlowDirectory:
        directory_id = self.__client.component.create_flow_directory(
            app_id=self._app_id, name=self._name, parent=self._parent
        )
        directory_dict = self.__client.component.get_flow_directory(
            app_id=self._app_id, directory_id=directory_id
        )

        directory = FlowDirectory.from_dict(data=directory_dict, client=self.__client)
        return directory


class FlowDirectory(BaseModel):
    app_id: str = Field(validation_alias="project_id")
    directory_id: str
    name: str
    __client: Client

    @property
    def type(self) -> DirectoryType:
        return DirectoryType.FLOW

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> NamedDirectoryDict:
        return {"directory_id": self.directory_id, "type": self.type, "name": self.name}

    def directories(self) -> NamedMapping[Self]:
        directory_dicts = self.__client.component.get_flow_directories_for_app(
            app_id=self.app_id, parent=self
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                self.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    def flows(self) -> NamedMapping[Flow]:
        flow_dicts = self.__client.component.get_flows_for_app(
            app_id=self.app_id, directory_id=self.directory_id
        )

        flows = {
            flow.flow_id: flow
            for flow in (
                Flow.from_dict(data=flow_dict, client=self.__client)
                for flow_dict in flow_dicts
            )
        }

        return NamedMapping(flows)



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from contextlib import ExitStack
from logging import Logger

import pandas as pd
import pytest
from ikigai.ikigai import Ikigai


def test_dataset_creation(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    datasets = app.datasets()
    assert len(datasets) == 0

    dataset = app.dataset.new(name=dataset_name).df(df1).build()
    cleanup.callback(dataset.delete)

    with pytest.raises(KeyError):
        datasets.get_id(dataset.dataset_id)

    datasets_after_creation = app.datasets()
    assert len(datasets_after_creation) == 1

    dataset_dict = dataset.to_dict()
    assert dataset_dict["name"] == dataset_name


def test_dataset_editing(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)
    dataset = app.dataset.new(name=dataset_name).df(df1).build()
    cleanup.callback(dataset.delete)

    dataset.rename(f"updated {dataset_name}")
    dataset.edit_data(df2)

    dataset_after_edit = app.datasets().get_id(dataset.dataset_id)
    round_trip_df2 = dataset_after_edit.df()

    assert dataset_after_edit.name == dataset.name
    assert dataset_after_edit.name == f"updated {dataset_name}"
    assert df2.columns.equals(round_trip_df2.columns)
    pd.testing.assert_frame_equal(
        df2, round_trip_df2, check_dtype=False, check_exact=False
    )


def test_dataset_download(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
    logger: Logger,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)
    dataset = app.dataset.new(name=dataset_name).df(df1).build()
    cleanup.callback(dataset.delete)

    round_trip_df1 = dataset.df()
    assert len(df1) == len(round_trip_df1)
    assert df1.columns.equals(round_trip_df1.columns)

    # v. helpful debug message when the test fails
    logger.info(
        "df1.dtypes:\n%r\n" "%r\n\n" "round_trip_df1.dtypes:\n%r\n" "%r\n\n",
        df1.dtypes,
        df1.head(),
        round_trip_df1.dtypes,
        round_trip_df1.head(),
    )

    pd.testing.assert_frame_equal(
        df1, round_trip_df1, check_dtype=False, check_exact=False
    )


def test_dataset_describe(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)
    dataset = app.dataset.new(name=dataset_name).df(df1).build()
    cleanup.callback(dataset.delete)

    description = dataset.describe()
    assert description is not None
    assert description["name"] == dataset_name
    assert description["project_id"] == app.app_id
    assert description["directory"] is not None
    assert description["directory"]["type"] == "DATASET"


def test_dataset_directories_creation(
    ikigai: Ikigai,
    app_name: str,
    dataset_directory_name_1: str,
    dataset_directory_name_2: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("App to test dataset directory creation")
        .build()
    )
    cleanup.callback(app.delete)

    assert len(app.dataset_directories()) == 0

    dataset_directory = app.dataset_directory.new(name=dataset_directory_name_1).build()
    assert len(app.dataset_directories()) == 1
    assert len(dataset_directory.directories()) == 0

    nested_dataset_directory = (
        app.dataset_directory.new(name=dataset_directory_name_2)
        .parent(dataset_directory)
        .build()
    )
    assert len(dataset_directory.directories()) == 1
    assert len(nested_dataset_directory.datasets()) == 0

    dataset_directories = app.dataset_directories()
    assert dataset_directories[dataset_directory_name_1]
    assert dataset_directories[dataset_directory_name_2]

    dataset = (
        app.dataset.new(name=dataset_name)
        .directory(directory=nested_dataset_directory)
        .df(df1)
        .build()
    )
    cleanup.callback(dataset.delete)
    assert len(nested_dataset_directory.datasets()) == 1
    assert len(dataset_directory.datasets()) == 0


"""
Regression Testing

- Each regression test should be of the format:
    f"test_{ticket_number}_{short_desc}"
"""


def test_iplt_7641_datasets(
    ikigai: Ikigai,
    app_name: str,
    dataset_directory_name_1: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("An app to test that app.flows gets all flows")
        .build()
    )
    cleanup.callback(app.delete)

    dataset_1 = app.dataset.new(name=dataset_name).df(df1).build()
    cleanup.callback(dataset_1.delete)

    dataset_directory = app.dataset_directory.new(name=dataset_directory_name_1).build()
    dataset_2 = (
        app.dataset.new(name=f"cloned-{dataset_name}")
        .directory(dataset_directory)
        .df(df1)
        .build()
    )
    cleanup.callback(dataset_2.delete)

    datasets = app.datasets()
    directory_datasets = dataset_directory.datasets()
    assert datasets
    assert directory_datasets
    assert len(directory_datasets) == 1
    assert len(datasets) >= len(directory_datasets)
    assert datasets[dataset_1.name]
    assert datasets[dataset_2.name]
    with pytest.raises(KeyError):
        directory_datasets[dataset_1.name]
    assert directory_datasets[dataset_2.name]



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any, Protocol, TypedDict

from ikigai.typing.protocol.directory import DirectoryDict
from ikigai.typing.protocol.generic import Empty


class ModelType(Protocol):
    @property
    def model_type(self) -> str: ...

    @property
    def sub_model_type(self) -> str: ...


class ModelDict(TypedDict):
    project_id: str
    model_id: str
    name: str
    latest_version_id: str
    directory: DirectoryDict
    model_type: str
    sub_model_type: str
    description: str
    created_at: str
    modified_at: str


class ModelVersionDict(TypedDict):
    version_id: str
    model_id: str
    version: str
    hyperparameters: dict
    metrics: dict
    created_at: str
    modified_at: str


class ModelSpecDict(TypedDict):
    name: str
    is_deprecated: bool
    is_hidden: bool
    keywords: list[str]
    metrics: ModelMetricsSpecDict
    sub_model_types: list[SubModelSpecDict]


class SubModelSpecDict(TypedDict):
    name: str
    is_deprecated: bool
    is_hidden: bool
    keywords: list[str]
    metrics: ModelMetricsSpecDict
    parameters: list[ModelParameterSpecDict]
    hyperparameters: list[ModelHyperparameterSpecDict]


ModelMetricsSpecDict = dict[str, Empty]


class ModelParameterSpecDict(TypedDict):
    name: str
    default_value: Any
    have_options: bool
    is_deprecated: bool
    is_hidden: bool
    is_list: bool
    options: list[Any]
    parameter_type: str


class ModelHyperparameterSpecDict(TypedDict):
    name: str
    default_value: Any
    have_options: bool
    have_sub_hyperparameters: bool
    hyperparameter_group: str
    hyperparameter_type: str
    is_deprecated: bool
    is_hidden: bool
    is_list: bool
    children: list[dict]
    options: list[Any]
    sub_hyperparameter_requirements: list[list]



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Protocol


class Named(Protocol):
    @property
    def name(self) -> str: ...


class Empty(Protocol):
    pass



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from enum import Enum
from typing import Protocol, TypedDict


class DirectoryType(str, Enum):
    APP = "PROJECT"
    DATASET = "DATASET"
    FLOW = "PIPELINE"
    MODEL = "MODEL"


class DirectoryDict(TypedDict):
    directory_id: str
    type: DirectoryType


class NamedDirectoryDict(DirectoryDict, TypedDict):
    name: str


class Directory(Protocol):
    @property
    def directory_id(self) -> str: ...

    @property
    def type(self) -> DirectoryType: ...

    def to_dict(self) -> DirectoryDict: ...



# SPDX-FileCopyrightText: 2025-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT


from contextlib import ExitStack

import pandas as pd
import pytest
from ikigai import Ikigai
from ikigai.components import FlowStatus


def test_flow_creation(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    flows = app.flows()
    assert len(flows) == 0

    flow = app.flow.new(name=flow_name).build()

    flow_dict = flow.to_dict()
    assert flow_dict["name"] == flow_name

    flows_after_creation = app.flows()
    assert len(flows_after_creation) == 1
    assert flows_after_creation[flow.name]

    flow.delete()
    flows_after_deletion = app.flows()
    assert len(flows_after_deletion) == 0
    with pytest.raises(KeyError):
        flows_after_deletion[flow.name]
    assert flow.flow_id not in flows_after_deletion


def test_flow_editing(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).build()
    cleanup.callback(flow.delete)
    # TODO: Update test once we can nicely create pipeline definitions

    flow.rename(f"updated {flow_name}")

    flow_after_edit = app.flows().get_id(flow.flow_id)

    assert flow_after_edit.name == flow.name
    assert flow_after_edit.name == f"updated {flow_name}"


def test_flow_status(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).build()
    cleanup.callback(flow.delete)

    status_report = flow.status()
    assert status_report.status == FlowStatus.IDLE
    assert status_report.progress is None
    assert not status_report.message


def test_flow_clone(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("App to test flow run").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()
    cleanup.callback(dataset.delete)

    # TODO: Update test once we can nicely create pipelines
    flow = (
        app.flow.new(name=flow_name)
        .definition(
            {
                "facets": [
                    {
                        "facet_id": "input",
                        "name": dataset.name,
                        "facet_uid": "I_005",  # Imported dataset
                        "arguments": {
                            "dataset_id": dataset.dataset_id,
                            "file_type": "csv",
                            "header": True,
                            "use_raw_file": False,
                        },
                    },
                    {
                        "facet_id": "count",
                        "name": "count",
                        "facet_uid": "M_003",  # Count
                        "arguments": {
                            "output_column_name": "count",
                            "sort": True,
                            "target_columns": df1.columns.to_list()[:-2],
                        },
                    },
                    {
                        "facet_id": "output",
                        "name": "output",
                        "facet_uid": "O_005",  # Exported dataset
                        "arguments": {
                            "dataset_name": f"output-{flow_name}",
                            "file_type": "csv",
                            "header": True,
                        },
                    },
                ],
                "arrows": [
                    {
                        "arguments": {},
                        "source": "input",
                        "destination": "count",
                    },
                    {
                        "arguments": {},
                        "source": "count",
                        "destination": "output",
                    },
                ],
            }
        )
        .build()
    )
    cleanup.callback(flow.delete)

    cloned_flow = (
        app.flow.new(name=f"clone of {flow_name}").definition(definition=flow).build()
    )
    cleanup.callback(cloned_flow.delete)

    flows = app.flows()

    # TODO: Add more assert to check that the cloning did happen
    assert flows[flow.name]
    assert flows[cloned_flow.name]


def test_flow_run_success(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("App to test flow run").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()
    cleanup.callback(dataset.delete)

    # TODO: Update test once we can nicely create pipelines
    flow = (
        app.flow.new(name=flow_name)
        .definition(
            {
                "facets": [
                    {
                        "facet_id": "input",
                        "name": dataset.name,
                        "facet_uid": "I_005",  # Imported dataset
                        "arguments": {
                            "dataset_id": dataset.dataset_id,
                            "file_type": "csv",
                            "header": True,
                            "use_raw_file": False,
                        },
                    },
                    {
                        "facet_id": "count",
                        "name": "count",
                        "facet_uid": "M_003",  # Count
                        "arguments": {
                            "output_column_name": "count",
                            "sort": True,
                            "target_columns": df1.columns.to_list()[:-2],
                        },
                    },
                    {
                        "facet_id": "output",
                        "name": "output",
                        "facet_uid": "O_005",  # Exported dataset
                        "arguments": {
                            "dataset_name": f"output-{flow_name}",
                            "file_type": "csv",
                            "header": True,
                        },
                    },
                ],
                "arrows": [
                    {
                        "arguments": {},
                        "source": "input",
                        "destination": "count",
                    },
                    {
                        "arguments": {},
                        "source": "count",
                        "destination": "output",
                    },
                ],
            }
        )
        .build()
    )
    cleanup.callback(flow.delete)

    log = flow.run()
    assert log.status == FlowStatus.SUCCESS
    assert log.erroneous_facet_id is None
    assert not log.data


def test_flow_run_fail(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("App to test flow run").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()
    cleanup.callback(dataset.delete)

    # TODO: Update test once we can nicely create pipelines
    flow = (
        app.flow.new(name=flow_name)
        .definition(
            {
                "facets": [
                    {
                        "facet_id": "input",
                        "name": dataset.name,
                        "facet_uid": "I_005",  # Imported dataset
                        "arguments": {
                            "dataset_id": dataset.dataset_id,
                            "file_type": "csv",
                            "header": True,
                            "use_raw_file": False,
                        },
                    },
                    {
                        "facet_id": "failing",
                        "name": "failing",
                        "facet_uid": "M_000",  # Python Code
                        "arguments": {
                            "script": (
                                "import pandas as pd\n"
                                "df = data\n"
                                "raise ValueError('Expected Error')\n"
                                "result = df\n"
                            ),
                        },
                    },
                    {
                        "facet_id": "output",
                        "name": "output",
                        "facet_uid": "O_005",  # Exported dataset
                        "arguments": {
                            "dataset_name": f"output-{flow_name}",
                            "file_type": "csv",
                            "header": True,
                        },
                    },
                ],
                "arrows": [
                    {
                        "arguments": {},
                        "source": "input",
                        "destination": "failing",
                    },
                    {
                        "arguments": {},
                        "source": "failing",
                        "destination": "output",
                    },
                ],
            }
        )
        .build()
    )
    cleanup.callback(flow.delete)

    log = flow.run()
    assert log.status == FlowStatus.FAILED
    assert log.erroneous_facet_id == "failing"
    assert log.data


def test_flow_directories_creation(
    ikigai: Ikigai,
    app_name: str,
    flow_directory_name_1: str,
    flow_directory_name_2: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("App to test flow directory creation")
        .build()
    )
    cleanup.callback(app.delete)

    assert len(app.flow_directories()) == 0

    flow_directory = app.flow_directory.new(name=flow_directory_name_1).build()
    assert len(app.flow_directories()) == 1
    assert len(flow_directory.directories()) == 0

    nested_flow_directory = (
        app.flow_directory.new(name=flow_directory_name_2)
        .parent(flow_directory)
        .build()
    )
    assert len(flow_directory.directories()) == 1
    assert len(nested_flow_directory.flows()) == 0

    flow_directories = app.flow_directories()
    assert flow_directories[flow_directory_name_1]
    assert flow_directories[flow_directory_name_2]

    flow = (
        app.flow.new(name=flow_name).directory(directory=nested_flow_directory).build()
    )
    cleanup.callback(flow.delete)
    assert len(nested_flow_directory.flows()) == 1
    assert len(flow_directory.flows()) == 0


"""
Regression Testing

- Each regression test should be of the format:
    f"test_{ticket_number}_{short_desc}"
"""


def test_iplt_7641_flows(
    ikigai: Ikigai,
    app_name: str,
    flow_directory_name_1: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("An app to test that app.flows gets all flows")
        .build()
    )
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).build()
    cleanup.callback(flow.delete)

    flow_directory = app.flow_directory.new(name=flow_directory_name_1).build()
    cloned_flow = (
        app.flow.new(name=f"cloned-{flow_name}")
        .directory(flow_directory)
        .definition(flow)
        .build()
    )
    cleanup.callback(cloned_flow.delete)

    flows = app.flows()
    directory_flows = flow_directory.flows()
    assert flows
    assert directory_flows
    assert len(directory_flows) == 1
    assert len(flows) >= len(directory_flows)
    assert flows[flow.name]
    assert flows[cloned_flow.name]
    with pytest.raises(KeyError):
        directory_flows[flow.name]
    assert directory_flows[cloned_flow.name]



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from dataclasses import InitVar
from functools import cache
from typing import Any, cast

from pydantic import Field
from pydantic.dataclasses import dataclass

from ikigai.client.session import Session
from ikigai.typing.api import (
    GetComponentsForProjectResponse,
    GetDatasetMultipartUploadUrlsResponse,
)
from ikigai.typing.protocol import (
    AppDict,
    DatasetDict,
    DatasetLogDict,
    Directory,
    DirectoryDict,
    FlowDefinitionDict,
    FlowDict,
    FlowLogDict,
    FlowStatusReportDict,
    ModelDict,
    ModelSpecDict,
    ModelType,
    ModelVersionDict,
)

_UNSET: Any = object()


@dataclass
class ComponentAPI:
    # Init only vars
    session: InitVar[Session]

    __session: Session = Field(init=False)

    def __post_init__(self, session: Session) -> None:
        self.__session = session

    def __hash__(self) -> int:
        # Enable the usage of @cache on specs related apis
        return hash(id(self))

    """
    App APIs
    """

    def create_app(
        self,
        name: str,
        description: str,
        directory: Directory | None,
    ) -> str:
        directory_dict = (
            cast(dict, directory.to_dict()) if directory is not None else {}
        )
        resp = self.__session.post(
            path="/component/create-project",
            json={
                "project": {
                    "name": name,
                    "description": description,
                    "directory": directory_dict,
                },
            },
        ).json()
        return resp["project_id"]

    def get_app(self, app_id: str) -> AppDict:
        app_dict = self.__session.get(
            path="/component/get-project", params={"project_id": app_id}
        ).json()["project"]

        return cast(AppDict, app_dict)

    def get_app_directories_for_user(
        self, directory_id: str = _UNSET
    ) -> list[DirectoryDict]:
        if directory_id == _UNSET:
            directory_id = ""

        directory_dicts = self.__session.get(
            path="/component/get-project-directories-for-user",
            params={"directory_id": directory_id},
        ).json()["directories"]

        return cast(list[DirectoryDict], directory_dicts)

    def get_apps_for_user(self, directory_id: str = _UNSET) -> list[AppDict]:
        fetch_all = directory_id == _UNSET
        if directory_id == _UNSET:
            directory_id = ""

        app_dicts = self.__session.get(
            path="/component/get-projects-for-user",
            params={"fetch_all": fetch_all, "directory_id": directory_id},
        ).json()["projects"]

        return cast(list[AppDict], app_dicts)

    def get_components_for_app(self, app_id: str) -> GetComponentsForProjectResponse:
        resp = self.__session.get(
            path="/component/get-components-for-project",
            params={"project_id": app_id},
        ).json()["project_components"][app_id]

        return cast(GetComponentsForProjectResponse, resp)

    def edit_app(
        self,
        app_id: str,
        name: str = _UNSET,
        directory: Directory = _UNSET,
        description: str = _UNSET,
    ) -> str:
        app: dict[str, Any] = {"project_id": app_id}

        if name != _UNSET:
            app["name"] = name
        if directory != _UNSET:
            app["directory"] = directory.to_dict()
        if description != _UNSET:
            app["description"] = description

        resp = self.__session.post(
            path="/component/edit-project",
            json={"project": app},
        ).json()

        return resp["project_id"]

    def delete_app(self, app_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-project",
            json={"project": {"project_id": app_id}},
        ).json()

        return resp["project_id"]

    """
    Dataset APIs
    """

    def create_dataset(
        self, app_id: str, name: str, directory: Directory | None
    ) -> str:
        directory_dict = (
            cast(dict, directory.to_dict()) if directory is not None else {}
        )
        resp = self.__session.post(
            path="/component/create-dataset",
            json={
                "dataset": {
                    "project_id": app_id,
                    "name": name,
                    "directory": directory_dict,
                },
            },
        ).json()
        return resp["dataset_id"]

    def get_dataset_download_url(self, app_id: str, dataset_id: str) -> str:
        resp = self.__session.get(
            path="/component/get-dataset-download-url",
            params={
                "project_id": app_id,
                "dataset_id": dataset_id,
            },
        ).json()
        return resp["url"]

    def get_dataset(self, app_id: str, dataset_id: str) -> DatasetDict:
        resp = self.__session.get(
            path="/component/get-dataset",
            params={"project_id": app_id, "dataset_id": dataset_id},
        ).json()
        dataset = resp["dataset"]

        return cast(DatasetDict, dataset)

    def get_datasets_for_app(
        self, app_id: str, directory_id: str = _UNSET
    ) -> list[DatasetDict]:
        params = {"project_id": app_id}
        if directory_id != _UNSET:
            params["directory_id"] = directory_id

        resp = self.__session.get(
            path="/component/get-datasets-for-project",
            params=params,
        ).json()
        datasets = resp["datasets"]

        return cast(list[DatasetDict], datasets)

    def get_dataset_multipart_upload_urls(
        self, dataset_id: str, app_id: str, filename: str, file_size: int
    ) -> GetDatasetMultipartUploadUrlsResponse:
        # TODO: Remove api-backwards compatibility once production apis are updated
        # Compatability: Also provide number of parts for backwards compatability
        CHUNK_SIZE = int(50e6)  # noqa: N806 -- 50 MB
        num_parts = (file_size + CHUNK_SIZE - 1) // CHUNK_SIZE
        resp = self.__session.get(
            path="/component/get-dataset-multipart-upload-urls",
            params={
                "dataset_id": dataset_id,
                "project_id": app_id,
                "filename": filename,
                "number_of_parts": num_parts,
                "file_size": file_size,
            },
        ).json()

        return GetDatasetMultipartUploadUrlsResponse(
            upload_id=resp["upload_id"],
            content_type=resp["content_type"],
            urls={
                int(chunk_idx): upload_url
                for chunk_idx, upload_url in resp["urls"].items()
            },
        )

    def get_dataset_log(
        self, app_id: str, dataset_id: str, limit: int = 5
    ) -> list[DatasetLogDict]:
        dataset_log = self.__session.get(
            path="/component/get-dataset-log",
            params={"dataset_id": dataset_id, "project_id": app_id, "limit": limit},
        ).json()["dataset_log"]

        return dataset_log

    def edit_dataset(
        self,
        app_id: str,
        dataset_id: str,
        name: str = _UNSET,
        directory: Directory = _UNSET,
    ) -> str:
        dataset: dict[str, Any] = {
            "project_id": app_id,
            "dataset_id": dataset_id,
        }

        if name != _UNSET:
            dataset["name"] = name
        if directory != _UNSET:
            dataset["directory"] = directory.to_dict()

        resp = self.__session.post(
            path="/component/edit-dataset",
            json={
                "dataset": dataset,
            },
        ).json()

        return resp["dataset_id"]

    def verify_dataset_upload(
        self, app_id: str, dataset_id: str, filename: str
    ) -> None:
        self.__session.get(
            path="/component/verify-dataset-upload",
            params={
                "project_id": app_id,
                "dataset_id": dataset_id,
                "filename": filename,
            },
        )
        return None

    def confirm_dataset_upload(self, app_id: str, dataset_id: str) -> str:
        resp = self.__session.get(
            path="/component/confirm-dataset-upload",
            params={"project_id": app_id, "dataset_id": dataset_id},
        ).json()
        return resp["status"]

    def abort_datset_multipart_upload(
        self, app_id: str, dataset_id: str, filename: str, upload_id: str
    ) -> None:
        self.__session.post(
            path="/component/complete-dataset-multipart-upload",
            json={
                "abort": True,
                "dataset": {
                    "dataset_id": dataset_id,
                    "project_id": app_id,
                    "filename": filename,
                },
                "upload_id": upload_id,
            },
        )
        return None

    def complete_datset_multipart_upload(
        self,
        app_id: str,
        dataset_id: str,
        filename: str,
        upload_id: str,
        etags: dict[int, str],
    ) -> None:
        self.__session.post(
            path="/component/complete-dataset-multipart-upload",
            json={
                "abort": False,
                "dataset": {
                    "dataset_id": dataset_id,
                    "project_id": app_id,
                    "filename": filename,
                },
                "upload_id": upload_id,
                "etags": etags,
            },
        )
        return None

    def delete_dataset(self, app_id: str, dataset_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-dataset",
            json={"dataset": {"project_id": app_id, "dataset_id": dataset_id}},
        ).json()

        return resp["dataset_id"]

    """
    Flow APIs
    """

    def create_flow(
        self,
        app_id: str,
        name: str,
        directory: Directory | None,
        flow_definition: FlowDefinitionDict,
    ) -> str:
        directory_dict = (
            cast(dict, directory.to_dict()) if directory is not None else {}
        )
        resp = self.__session.post(
            path="/component/create-pipeline",
            json={
                "pipeline": {
                    "project_id": app_id,
                    "name": name,
                    "directory": directory_dict,
                    "definition": flow_definition,
                },
            },
        ).json()
        return resp["pipeline_id"]

    def get_flow(self, flow_id: str) -> FlowDict:
        flow = self.__session.get(
            path="/component/get-pipeline", params={"pipeline_id": flow_id}
        ).json()["pipeline"]

        return cast(FlowDict, flow)

    def get_flows_for_app(
        self, app_id: str, directory_id: str = _UNSET
    ) -> list[FlowDict]:
        params = {"project_id": app_id}

        if directory_id != _UNSET:
            params["directory_id"] = directory_id

        flows = self.__session.get(
            path="/component/get-pipelines-for-project",
            params=params,
        ).json()["pipelines"]

        return cast(list[FlowDict], flows)

    def get_flow_log(
        self, app_id: str, flow_id: str, max_count: int
    ) -> list[FlowLogDict]:
        log_dicts = self.__session.get(
            path="/component/get-pipeline-log",
            params={
                "pipeline_id": flow_id,
                "project_id": app_id,
                "limit": max_count,
            },
        ).json()["pipeline_log"]

        return log_dicts

    def edit_flow(
        self,
        app_id: str,
        flow_id: str,
        name: str | None = None,
        directory: Directory | None = None,
        flow_definition: FlowDefinitionDict | None = None,
    ) -> str:
        pipeline: dict[str, Any] = {
            "project_id": app_id,
            "pipeline_id": flow_id,
        }

        if name is not None:
            pipeline["name"] = name
        if directory is not None:
            pipeline["directory"] = directory.to_dict()
        if flow_definition is not None:
            pipeline["definition"] = flow_definition

        resp = self.__session.post(
            path="/component/edit-pipeline", json={"pipeline": pipeline}
        ).json()
        return resp["pipeline_id"]

    def delete_flow(self, app_id: str, flow_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-pipeline",
            json={"pipeline": {"project_id": app_id, "pipeline_id": flow_id}},
        ).json()

        return resp["pipeline_id"]

    def run_flow(self, app_id: str, flow_id: str) -> str:
        resp = self.__session.post(
            path="/component/run-pipeline",
            json={"pipeline": {"project_id": app_id, "pipeline_id": flow_id}},
        ).json()

        return resp["pipeline_id"]

    def is_flow_runing(self, app_id: str, flow_id: str) -> FlowStatusReportDict:
        resp = self.__session.get(
            path="/component/is-pipeline-running",
            params={"project_id": app_id, "pipeline_id": flow_id},
        ).json()

        # BE is a bit inconsistent with the response so clean it up
        status = resp["progress"]["status"] if resp["status"] else "IDLE"
        progress = resp["progress"].get("progress")
        message = resp["progress"].get("message")

        return FlowStatusReportDict(
            status=status,
            progress=progress,
            message=message,
        )

    """
    Model APIs
    """

    def create_model(
        self,
        app_id: str,
        name: str,
        directory: Directory | None,
        model_type: ModelType,
        description: str,
    ) -> str:
        directory_dict = (
            cast(dict, directory.to_dict()) if directory is not None else {}
        )

        resp = self.__session.post(
            path="/component/create-model",
            json={
                "model": {
                    "project_id": app_id,
                    "name": name,
                    "directory": directory_dict,
                    "model_type": model_type.model_type,
                    "sub_model_type": model_type.sub_model_type,
                    "description": description,
                }
            },
        ).json()

        return resp["model_id"]

    def get_model(self, app_id: str, model_id: str) -> ModelDict:
        resp = self.__session.get(
            path="/component/get-model",
            params={"project_id": app_id, "model_id": model_id},
        ).json()
        model = resp["model"]
        return cast(ModelDict, model)

    def get_models_for_app(
        self, app_id: str, directory_id: str = _UNSET
    ) -> list[ModelDict]:
        params = {"project_id": app_id}
        if directory_id != _UNSET:
            params["directory_id"] = directory_id

        resp = self.__session.get(
            path="/component/get-models-for-project",
            params=params,
        ).json()
        models = resp["models"]

        return cast(list[ModelDict], models)

    @cache
    def get_model_specs(self) -> list[ModelSpecDict]:
        resp = self.__session.get(
            path="/component/get-model-specs",
        ).json()

        model_specs = resp.values()

        return cast(list[ModelSpecDict], model_specs)

    def edit_model(
        self,
        app_id: str,
        model_id: str,
        name: str = _UNSET,
        directory: Directory = _UNSET,
        description: str = _UNSET,
    ) -> str:
        model: dict[str, Any] = {
            "project_id": app_id,
            "model_id": model_id,
        }

        if name != _UNSET:
            model["name"] = name
        if directory != _UNSET:
            model["directory"] = directory.to_dict()
        if description != _UNSET:
            model["description"] = description

        resp = self.__session.post(
            path="/component/edit-model",
            json={"model": model},
        ).json()

        return resp["model_id"]

    def delete_model(self, app_id: str, model_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-model",
            json={"model": {"project_id": app_id, "model_id": model_id}},
        ).json()

        return resp["model_id"]

    """
    Model Version APIs
    """

    def get_model_version(self, app_id: str, version_id: str) -> ModelVersionDict:
        resp = self.__session.get(
            path="/component/get-model-version",
            params={"project_id": app_id, "version_id": version_id},
        ).json()
        model_version = resp["model_version"]

        return cast(ModelVersionDict, model_version)

    def get_model_versions(self, app_id: str, model_id: str) -> list[ModelVersionDict]:
        resp = self.__session.get(
            path="/component/get-model-versions",
            params={"project_id": app_id, "model_id": model_id},
        ).json()
        model_versions = resp["versions"]

        return cast(list[ModelVersionDict], model_versions)

    """
    Directory APIs
    """

    def create_dataset_directory(
        self, app_id: str, name: str, parent: Directory | None = None
    ) -> str:
        parent_id = parent.directory_id if parent else ""

        resp = self.__session.post(
            path="/component/create-dataset-directory",
            json={
                "directory": {
                    "name": name,
                    "project_id": app_id,
                    "parent_id": parent_id,
                }
            },
        ).json()

        return resp["directory_id"]

    def get_dataset_directory(self, app_id: str, directory_id: str) -> DirectoryDict:
        directory = self.__session.get(
            path="/component/get-dataset-directory",
            params={"project_id": app_id, "directory_id": directory_id},
        ).json()["directory"]

        return cast(DirectoryDict, directory)

    def get_dataset_directories_for_app(
        self, app_id: str, parent: Directory = _UNSET
    ) -> list[DirectoryDict]:
        params = {"project_id": app_id}
        if parent != _UNSET:
            params["directory_id"] = parent.directory_id

        resp = self.__session.get(
            path="/component/get-dataset-directories-for-project",
            params=params,
        ).json()
        directories = resp["directories"]

        return cast(list[DirectoryDict], directories)

    def create_flow_directory(
        self, app_id: str, name: str, parent: Directory | None = None
    ) -> str:
        parent_id = parent.directory_id if parent else ""

        resp = self.__session.post(
            path="/component/create-pipeline-directory",
            json={
                "directory": {
                    "name": name,
                    "project_id": app_id,
                    "parent_id": parent_id,
                }
            },
        ).json()

        return resp["directory_id"]

    def get_flow_directory(self, app_id: str, directory_id: str) -> DirectoryDict:
        directory = self.__session.get(
            path="/component/get-pipeline-directory",
            params={"project_id": app_id, "directory_id": directory_id},
        ).json()["directory"]

        return cast(DirectoryDict, directory)

    def get_flow_directories_for_app(
        self, app_id: str, parent: Directory = _UNSET
    ) -> list[DirectoryDict]:
        params = {"project_id": app_id}
        if parent != _UNSET:
            params["directory_id"] = parent.directory_id

        resp = self.__session.get(
            path="/component/get-pipeline-directories-for-project",
            params=params,
        ).json()

        directories = resp["directories"]
        return cast(list[DirectoryDict], directories)

    def create_model_directory(
        self, app_id: str, name: str, parent: Directory | None = None
    ) -> str:
        parent_id = parent.directory_id if parent else ""

        resp = self.__session.post(
            path="/component/create-model-directory",
            json={
                "directory": {
                    "name": name,
                    "project_id": app_id,
                    "parent_id": parent_id,
                }
            },
        ).json()

        return resp["directory_id"]

    def get_model_directory(self, app_id: str, directory_id: str) -> DirectoryDict:
        directory = self.__session.get(
            path="/component/get-model-directory",
            params={"project_id": app_id, "directory_id": directory_id},
        ).json()["directory"]

        return cast(DirectoryDict, directory)

    def get_model_directories_for_app(
        self, app_id: str, parent: Directory = _UNSET
    ) -> list[DirectoryDict]:
        params = {"project_id": app_id}
        if parent != _UNSET:
            params["directory_id"] = parent.directory_id

        resp = self.__session.get(
            path="/component/get-model-directories-for-project",
            params=params,
        ).json()

        directories = resp["directories"]
        return cast(list[DirectoryDict], directories)



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.components.app import App, AppBuilder, AppDirectory
from ikigai.components.dataset import (
    Dataset,
    DatasetBuilder,
    DatasetDirectory,
    DatasetDirectoryBuilder,
)
from ikigai.components.flow import (
    Flow,
    FlowBuilder,
    FlowDirectory,
    FlowDirectoryBuilder,
    FlowStatus,
)
from ikigai.components.model import (
    Model,
    ModelBuilder,
    ModelDirectory,
    ModelDirectoryBuilder,
)

__all__ = [
    "App",
    "AppBuilder",
    "AppDirectory",
    "Dataset",
    "DatasetBuilder",
    "DatasetDirectory",
    "DatasetDirectoryBuilder",
    "Flow",
    "FlowBuilder",
    "FlowDirectoryBuilder",
    "FlowStatus",
    "FlowDirectory",
    "Model",
    "ModelBuilder",
    "ModelDirectory",
    "ModelDirectoryBuilder",
]



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime
from typing import Any

from pydantic import BaseModel, EmailStr, Field

from ikigai import components
from ikigai.client import Client
from ikigai.typing.protocol import Directory, DirectoryType, NamedDirectoryDict
from ikigai.utils.compatibility import Self
from ikigai.utils.named_mapping import NamedMapping


class AppBuilder:
    _name: str
    _description: str
    _directory: Directory | None
    _icon: str
    _images: list[str]
    __client: Client

    def __init__(self, client: Client) -> None:
        self.__client = client
        self._name = ""
        self._description = ""
        self._directory = None
        self._icon = ""
        self._images = []

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def description(self, description: str) -> Self:
        self._description = description
        return self

    def directory(self, directory: Directory) -> Self:
        self._directory = directory
        return self

    def build(self) -> App:
        app_id = self.__client.component.create_app(
            name=self._name,
            description=self._description,
            directory=self._directory,
        )
        app_dict = self.__client.component.get_app(app_id=app_id)
        app = App.from_dict(data=app_dict, client=self.__client)
        return app


class App(BaseModel):
    app_id: str = Field(validation_alias="project_id")
    name: str
    owner: EmailStr
    description: str
    created_at: datetime
    modified_at: datetime
    last_used_at: datetime
    __client: Client

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        self = cls.model_validate(data)
        self.__client = client
        return self

    """
    Operations on App
    """

    def to_dict(self) -> dict:
        return {
            "app_id": self.app_id,
            "name": self.name,
            "owner": self.owner,
            "description": self.description,
            "created_at": self.created_at,
            "modified_at": self.modified_at,
            "last_used_at": self.last_used_at,
        }

    def delete(self) -> None:
        self.__client.component.delete_app(app_id=self.app_id)
        return None

    def rename(self, name: str) -> Self:
        _ = self.__client.component.edit_app(app_id=self.app_id, name=name)
        # TODO: handle error case, currently it is a raise NotImplemented from Session
        self.name = name
        return self

    def move(self, directory: Directory) -> Self:
        _ = self.__client.component.edit_app(app_id=self.app_id, directory=directory)
        return self

    def update_description(self, description: str) -> Self:
        _ = self.__client.component.edit_app(
            app_id=self.app_id, description=description
        )
        # TODO: handle error case, currently it is a raise NotImplemented from Session
        self.description = description
        return self

    def describe(self) -> dict[str, Any]:
        components = self.__client.component.get_components_for_app(app_id=self.app_id)

        # Combine components information with app info
        return_value = {
            "app": self.to_dict(),
            "components": components,
        }

        return return_value

    """
    Access Components in the App
    """

    def datasets(self) -> NamedMapping[components.Dataset]:
        dataset_dicts = self.__client.component.get_datasets_for_app(app_id=self.app_id)
        datasets = {
            dataset.dataset_id: dataset
            for dataset in (
                components.Dataset.from_dict(data=dataset_dict, client=self.__client)
                for dataset_dict in dataset_dicts
            )
        }

        return NamedMapping(datasets)

    @property
    def dataset(self) -> components.DatasetBuilder:
        return components.DatasetBuilder(client=self.__client, app_id=self.app_id)

    def dataset_directories(self) -> NamedMapping[components.DatasetDirectory]:
        directory_dicts = self.__client.component.get_dataset_directories_for_app(
            app_id=self.app_id
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                components.DatasetDirectory.from_dict(
                    data=directory_dict, client=self.__client
                )
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    @property
    def dataset_directory(self) -> components.DatasetDirectoryBuilder:
        return components.DatasetDirectoryBuilder(
            client=self.__client, app_id=self.app_id
        )

    def flows(self) -> NamedMapping[components.Flow]:
        flow_dicts = self.__client.component.get_flows_for_app(app_id=self.app_id)

        flows = {
            flow.flow_id: flow
            for flow in (
                components.Flow.from_dict(data=flow_dict, client=self.__client)
                for flow_dict in flow_dicts
            )
        }

        return NamedMapping(flows)

    @property
    def flow(self) -> components.FlowBuilder:
        return components.FlowBuilder(client=self.__client, app_id=self.app_id)

    def flow_directories(self) -> NamedMapping[components.FlowDirectory]:
        directory_dicts = self.__client.component.get_flow_directories_for_app(
            app_id=self.app_id
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                components.FlowDirectory.from_dict(
                    data=directory_dict, client=self.__client
                )
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    @property
    def flow_directory(self) -> components.FlowDirectoryBuilder:
        return components.FlowDirectoryBuilder(client=self.__client, app_id=self.app_id)

    def models(self) -> NamedMapping[components.Model]:
        model_dicts = self.__client.component.get_models_for_app(app_id=self.app_id)

        models = {
            model.model_id: model
            for model in (
                components.Model.from_dict(data=model_dict, client=self.__client)
                for model_dict in model_dicts
            )
        }

        return NamedMapping(models)

    @property
    def model(self) -> components.ModelBuilder:
        return components.ModelBuilder(client=self.__client, app_id=self.app_id)

    def model_directories(self) -> NamedMapping[components.ModelDirectory]:
        directory_dicts = self.__client.component.get_model_directories_for_app(
            app_id=self.app_id
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                components.ModelDirectory.from_dict(
                    data=directory_dict, client=self.__client
                )
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    @property
    def model_directory(self) -> components.ModelDirectoryBuilder:
        return components.ModelDirectoryBuilder(
            client=self.__client, app_id=self.app_id
        )


class AppDirectory(BaseModel):
    directory_id: str
    name: str
    created_at: datetime
    modified_at: datetime
    __client: Client

    @property
    def type(self) -> DirectoryType:
        return DirectoryType.APP

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> NamedDirectoryDict:
        return {"directory_id": self.directory_id, "type": self.type, "name": self.name}

    def directories(self) -> NamedMapping[Self]:
        directory_dicts = self.__client.component.get_app_directories_for_user(
            directory_id=self.directory_id,
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                self.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    def apps(self) -> NamedMapping[App]:
        app_dicts = self.__client.component.get_apps_for_user(
            directory_id=self.directory_id
        )

        apps = {
            app.app_id: app
            for app in (
                App.from_dict(data=app_dict, client=self.__client)
                for app_dict in app_dicts
            )
        }

        return NamedMapping(apps)



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import io
import logging
import math
import time
from collections.abc import Mapping
from datetime import datetime
from enum import Enum
from http import HTTPStatus
from pathlib import Path
from typing import Any

import pandas as pd
import requests
from pydantic import BaseModel, Field

from ikigai.client import Client
from ikigai.typing.protocol import (
    Directory,
    DirectoryType,
    NamedDirectoryDict,
)
from ikigai.utils.compatibility import Self
from ikigai.utils.named_mapping import NamedMapping

logger = logging.getLogger("ikigai.components")


def __upload_data(
    client: Client,
    app_id: str,
    dataset_id: str,
    data: bytes,
    filename: str,
) -> None:
    file_size = len(data)
    multipart_upload_metadata = client.component.get_dataset_multipart_upload_urls(
        dataset_id=dataset_id,
        app_id=app_id,
        filename=filename,
        file_size=file_size,
    )

    content_type = multipart_upload_metadata["content_type"]
    upload_urls = multipart_upload_metadata["urls"]
    upload_id = multipart_upload_metadata["upload_id"]

    num_chunks = len(upload_urls)
    chunk_size = math.ceil(file_size / num_chunks)

    etags: dict[int, str] = {}
    try:
        with requests.session() as request:
            request.headers.update(
                {"Content-Type": content_type, "Cache-Control": "no-cache"}
            )
            for idx, (chunk_idx, upload_url) in enumerate(sorted(upload_urls.items())):
                chunk_start, chunk_end = (idx * chunk_size, (idx + 1) * chunk_size)
                chunk = data[chunk_start : min(chunk_end, file_size)]
                resp = request.put(url=upload_url, data=chunk)
                assert resp.status_code == HTTPStatus.OK

                # Get etags from response header
                etags[chunk_idx] = resp.headers["ETag"]
    except Exception:
        client.component.abort_datset_multipart_upload(
            app_id=app_id,
            dataset_id=dataset_id,
            filename=filename,
            upload_id=upload_id,
        )
        raise

    # Complete Dataset upload
    client.component.complete_datset_multipart_upload(
        app_id=app_id,
        dataset_id=dataset_id,
        filename=filename,
        upload_id=upload_id,
        etags=etags,
    )


def _upload_data(
    client: Client, app_id: str, dataset_id: str, name: str, data: bytes
) -> None:
    assert data is not None
    filename = f"{name}.csv"

    __upload_data(
        client=client,
        app_id=app_id,
        dataset_id=dataset_id,
        data=data,
        filename=filename,
    )

    upload_completion_time = time.time()
    client.component.verify_dataset_upload(
        app_id=app_id, dataset_id=dataset_id, filename=filename
    )

    dataset_status: str = "RUNNING"
    while dataset_status == "RUNNING":
        # Block thread while dataset is still being processed
        time.sleep(0.25)
        dataset_status = client.component.confirm_dataset_upload(
            app_id=app_id, dataset_id=dataset_id
        )

    if dataset_status != "SUCCESS":
        error_msg = f"Dataset upload failed, upload ended with status {dataset_status}"
        raise RuntimeError(error_msg)

    dataset_logs = client.component.get_dataset_log(
        app_id=app_id, dataset_id=dataset_id, limit=5
    )
    dataset_verified = [
        log
        for log in dataset_logs
        if log["status"] == "SUCCESS"
        and int(log["timestamp"]) > upload_completion_time
        and log["job_type"] == "UPLOAD_DATASET"
    ]
    if not dataset_verified:
        # TODO: Improve Error
        error_msg = "Dataset failed to verify"
        raise RuntimeError(error_msg, dataset_logs)


class DatasetBuilder:
    _app_id: str
    _name: str
    _data: bytes | None
    _directory: Directory | None
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._data = None
        self._directory = None

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def df(self, data: pd.DataFrame) -> Self:
        buffer = io.BytesIO()
        data.to_csv(buffer, index_label=False, index=False)
        self._data = buffer.getvalue()
        return self

    def csv(self, file: Path | str) -> Self:
        if isinstance(file, str):
            file = Path(file)
        with file.open("rb") as fp:
            self._data = fp.read()
        return self

    def directory(self, directory: Directory) -> Self:
        self._directory = directory
        return self

    def build(self) -> Dataset:
        if self._data is None:
            error_msg = "Dataset is empty"
            raise ValueError(error_msg)

        dataset_id = self.__client.component.create_dataset(
            app_id=self._app_id, name=self._name, directory=self._directory
        )

        try:
            _upload_data(
                client=self.__client,
                app_id=self._app_id,
                dataset_id=dataset_id,
                name=self._name,
                data=self._data,
            )
        except Exception:
            # Delete created record and re-raise
            self.__client.component.delete_dataset(
                app_id=self._app_id, dataset_id=dataset_id
            )
            raise

        # Populate Dataset object
        dataset_dict = self.__client.component.get_dataset(
            app_id=self._app_id, dataset_id=dataset_id
        )
        dataset = Dataset.from_dict(data=dataset_dict, client=self.__client)
        return dataset


class DataType(str, Enum):
    NUMERIC = "NUMERIC"
    TEXT = "TEXT"
    CATEGORICAL = "CATEGORICAL"
    TIME = "TIME"


class ColumnDataType(BaseModel):
    data_type: DataType
    data_formats: dict[str, str]


class Dataset(BaseModel):
    app_id: str = Field(validation_alias="project_id")
    dataset_id: str
    name: str
    filename: str
    file_extension: str
    data_types: dict[str, ColumnDataType]
    size: int
    created_at: datetime
    modified_at: datetime
    __client: Client

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> dict:
        return {
            "dataset_id": self.dataset_id,
            "name": self.name,
            "filename": self.filename,
            "file_extension": self.file_extension,
            "data_types": self.data_types,
            "size": self.size,
            "created_at": self.created_at,
            "modified_at": self.modified_at,
        }

    def delete(self) -> None:
        self.__client.component.delete_dataset(
            app_id=self.app_id, dataset_id=self.dataset_id
        )
        return None

    def rename(self, name: str) -> Self:
        _ = self.__client.component.edit_dataset(
            app_id=self.app_id, dataset_id=self.dataset_id, name=name
        )
        # TODO: handle error case, currently it is a raise NotImplemented from Session
        self.name = name
        return self

    def move(self, directory: Directory) -> Self:
        _ = self.__client.component.edit_dataset(
            app_id=self.app_id,
            dataset_id=self.dataset_id,
            directory=directory,
        )
        return self

    def df(self, **parser_options) -> pd.DataFrame:
        download_url = self.__client.component.get_dataset_download_url(
            app_id=self.app_id,
            dataset_id=self.dataset_id,
        )
        return pd.read_csv(download_url, **parser_options)

    def edit_data(self, data: pd.DataFrame) -> None:
        buffer = io.BytesIO()
        data.to_csv(buffer, index_label=False, index=False)

        _upload_data(
            client=self.__client,
            app_id=self.app_id,
            dataset_id=self.dataset_id,
            name=self.name,
            data=buffer.getvalue(),
        )

    def describe(self) -> Mapping[str, Any]:
        dataset = self.__client.component.get_dataset(
            app_id=self.app_id, dataset_id=self.dataset_id
        )
        return dataset


class DatasetDirectoryBuilder:
    _app_id: str
    _name: str
    _parent: Directory | None
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._parent = None

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def parent(self, parent: Directory) -> Self:
        self._parent = parent
        return self

    def build(self) -> DatasetDirectory:
        directory_id = self.__client.component.create_dataset_directory(
            app_id=self._app_id, name=self._name, parent=self._parent
        )
        directory_dict = self.__client.component.get_dataset_directory(
            app_id=self._app_id, directory_id=directory_id
        )

        directory = DatasetDirectory.from_dict(
            data=directory_dict, client=self.__client
        )
        return directory


class DatasetDirectory(BaseModel):
    app_id: str = Field(validation_alias="project_id")
    directory_id: str
    name: str
    __client: Client

    @property
    def type(self) -> DirectoryType:
        return DirectoryType.DATASET

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> NamedDirectoryDict:
        return {"directory_id": self.directory_id, "type": self.type, "name": self.name}

    def directories(self) -> NamedMapping[Self]:
        directory_dicts = self.__client.component.get_dataset_directories_for_app(
            app_id=self.app_id, parent=self
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                self.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    def datasets(self) -> NamedMapping[Dataset]:
        dataset_dicts = self.__client.component.get_datasets_for_app(
            app_id=self.app_id,
            directory_id=self.directory_id,
        )
        datasets = {
            dataset.dataset_id: dataset
            for dataset in (
                Dataset.from_dict(data=dataset_dict, client=self.__client)
                for dataset_dict in dataset_dicts
            )
        }

        return NamedMapping(datasets)



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from pydantic import AnyUrl, EmailStr, Field
from pydantic.dataclasses import dataclass

from ikigai import components
from ikigai.client import Client
from ikigai.utils.named_mapping import NamedMapping


@dataclass
class Ikigai:
    user_email: EmailStr
    api_key: str = Field(repr=False)
    base_url: AnyUrl = Field(default=AnyUrl("https://api.ikigailabs.io"))
    __client: Client = Field(init=False)

    def __post_init__(self) -> None:
        self.__client = Client(
            user_email=self.user_email, api_key=self.api_key, base_url=self.base_url
        )

    def apps(self) -> NamedMapping[components.App]:
        app_dicts = self.__client.component.get_apps_for_user()
        apps = {
            app.app_id: app
            for app in (
                components.App.from_dict(data=app_dict, client=self.__client)
                for app_dict in app_dicts
            )
        }

        return NamedMapping(apps)

    @property
    def app(self) -> components.AppBuilder:
        return components.AppBuilder(client=self.__client)

    def directories(self) -> NamedMapping[components.AppDirectory]:
        directory_dicts = self.__client.component.get_app_directories_for_user()
        directories = {
            directory.directory_id: directory
            for directory in (
                components.AppDirectory.from_dict(
                    data=directory_dict, client=self.__client
                )
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TypedDict

from ikigai.typing.protocol.directory import DirectoryDict


class DatasetDict(TypedDict):
    project_id: str
    dataset_id: str
    name: str
    filename: str
    data_types: dict[str, DataTypeDict]
    directory: DirectoryDict
    is_optimized: bool
    file_extension: str
    size: int
    is_visible: bool
    created_at: str
    modified_at: str


class DataTypeDict(TypedDict):
    data_type: str
    data_formats: str


class DatasetLogDict(TypedDict):
    status: str
    timestamp: str
    job_type: str



# SPDX-FileCopyrightText: 2025-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from contextlib import ExitStack

import pytest
from ikigai import Ikigai


def test_model_types(
    ikigai: Ikigai,
    app_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    model_types = app.model.model_types
    assert model_types is not None
    assert len(model_types) > 0
    lasso = model_types["Linear"]["Lasso"]

    assert lasso is not None
    assert lasso.model_type == "Linear"
    assert lasso.sub_model_type == "Lasso"


def test_model_creation(
    ikigai: Ikigai,
    app_name: str,
    model_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    models = app.models()
    assert len(models) == 0

    model_types = app.model.model_types
    model = (
        app.model.new(model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )

    models_after_creation = app.models()
    assert len(models_after_creation) == 1
    assert models_after_creation[model.name]

    model.delete()
    models_after_deletion = app.models()
    assert len(models_after_deletion) == 0

    with pytest.raises(KeyError):
        models_after_deletion[model.name]
    assert model.model_id not in models_after_deletion


def test_model_editing(
    ikigai: Ikigai,
    app_name: str,
    model_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    model_types = app.model.model_types
    model = (
        app.model.new(model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )
    cleanup.callback(model.delete)

    model.rename(f"updated {model_name}")
    model.update_description("updated description")

    model_after_edit = app.models().get_id(model.model_id)
    assert model_after_edit.name == model.name
    assert model_after_edit.name == f"updated {model_name}"
    assert model_after_edit.description == model.description
    assert model_after_edit.description == "updated description"
    assert model_after_edit.model_type == model.model_type
    assert model_after_edit.model_type == "Linear"
    assert model_after_edit.sub_model_type == model.sub_model_type
    assert model_after_edit.sub_model_type == "Lasso"


def test_model_describe(
    ikigai: Ikigai,
    app_name: str,
    model_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    model_types = app.model.model_types
    model = (
        app.model.new(model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )
    cleanup.callback(model.delete)

    model_description = model.describe()
    assert model_description is not None
    assert model_description["name"] == model.name
    assert model_description["description"] == model.description



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Mapping
from typing import Any, TypedDict


class GetDatasetMultipartUploadUrlsResponse(TypedDict):
    upload_id: str
    content_type: str
    urls: dict[int, str]


class GetComponentsForProjectResponse(TypedDict):
    charts: list[Mapping[str, Any]]
    connectors: list[Mapping[str, Any]]
    dashboards: list[Mapping[str, Any]]
    datasets: list[Mapping[str, Any]]
    databases: list[Mapping[str, Any]]
    pipelines: list[Mapping[str, Any]]
    models: list[Mapping[str, Any]]
    external_resources: list[Mapping[str, Any]]
    users: list[Mapping[str, Any]]
    connector_directories: list[Mapping[str, Any]]
    dashboard_directories: list[Mapping[str, Any]]
    dataset_directories: list[Mapping[str, Any]]
    database_directories: list[Mapping[str, Any]]
    pipeline_directories: list[Mapping[str, Any]]
    model_directories: list[Mapping[str, Any]]
    external_resource_directories: list[Mapping[str, Any]]



# SPDX-FileCopyrightText: 2024-present Harsh Parekh <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.typing.protocol.app import AppDict
from ikigai.typing.protocol.dataset import DatasetDict, DatasetLogDict
from ikigai.typing.protocol.directory import (
    Directory,
    DirectoryDict,
    DirectoryType,
    NamedDirectoryDict,
)
from ikigai.typing.protocol.flow import (
    ArrowDict,
    FacetDict,
    FlowDefinitionDict,
    FlowDict,
    FlowLogDict,
    FlowModelVariableDict,
    FlowStatusReportDict,
    FlowVariableDict,
)
from ikigai.typing.protocol.generic import Named
from ikigai.typing.protocol.model import (
    ModelDict,
    ModelSpecDict,
    ModelType,
    ModelVersionDict,
)

__all__ = [
    # App Protocol
    "AppDict",
    # Dataset Protocol
    "DatasetDict",
    "DatasetLogDict",
    # Directory Protocol
    "Directory",
    "DirectoryDict",
    "DirectoryType",
    "NamedDirectoryDict",
    # Flow Protocol
    "ArrowDict",
    "FacetDict",
    "FlowDict",
    "FlowLogDict",
    "FlowDefinitionDict",
    "FlowModelVariableDict",
    "FlowStatusReportDict",
    "FlowVariableDict",
    # Model Protocol
    "ModelDict",
    "ModelSpecDict",
    "ModelType",
    "ModelVersionDict",
    # Generic Protocol
    "Named",
]
